//-----------------------------------------------------------------------
// File : SampleRenderer.cpp
// Desc : Sample of Renderer implementation.
// Copyright(c) Project Asura. All right reserved.
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// History
//-----------------------------------------------------------------------
// <---> 2011/08/14 [Pocol] VK쐬
// <End>


//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include "SampleRenderer.h"


//-----------------------------------------------------------------------
// Using Statements
//-----------------------------------------------------------------------
using namespace Asura;
using namespace Asura::SampleProgram;
using namespace Asura::Graphics;
using namespace Asura::Math;
using namespace Asura::Util;


namespace Asura {
namespace SampleProgram {

/////////////////////////////////////////////////////////////////////////
// SampleRenderer class
/////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------
//! @brief lݒ
//-----------------------------------------------------------------------
SampleRenderer::SampleRenderer()
{
	D3DXMatrixIdentity( &mWorld );
	D3DXMatrixIdentity( &mView );
	D3DXMatrixIdentity( &mProjection );

	mViewDir  = D3DXVECTOR3( 0.0f, 0.0f, 1.0f );
	mLightDir = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );

    for( uint32_t i=0; i<2; ++i )
    {
    	mpSRV[ i ] = null;
    	mpSmp[ i ] = null;
        mpTex[ i ] = null;
        mpRTV[ i ] = null;
    }
    
	mpQuadVB = null;
}

//-----------------------------------------------------------------------
//! @brief j
//-----------------------------------------------------------------------
SampleRenderer::~SampleRenderer()
{
	/* DO_NOTHING */
}

//-----------------------------------------------------------------------
//! @brief 
//-----------------------------------------------------------------------
bool SampleRenderer::OnInit()
{
	// Jݒ
	D3DXVECTOR3 cameraPosition = D3DXVECTOR3( 0.0f, 2.5f, -5.0f );
	D3DXVECTOR3 cameraTarget   = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	D3DXVECTOR3 cameraUpVector = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );

	// svZ
	D3DXMatrixIdentity( &mWorld );
	D3DXMatrixLookAtLH( &mView, &cameraPosition, &cameraTarget, &cameraUpVector );
	D3DXMatrixPerspectiveFovLH( &mProjection, Math::PiOver4, DemoApp::GetInstance()->GetAspectRatio(), 1.0f, 1000.0f );

	// xNgݒ
	mViewDir = cameraTarget - cameraPosition;
	mLightDir = D3DXVECTOR3( -0.5f, 0.25f, -1.0f );

	// f[h
	if ( !mSaturn.LoadFromX( "Media/Models/saturn.x" ) )
	{
		ELOG( "Error : Model Load Failed." );
		return false;
	}
	if ( !mGround.LoadFromX( "Media/Models/ground.x" ) )
	{
		ELOG( "Error : Model Load Failed." );
		return false;
	}
   
    const float dispersion_sq = 5.0f;
    ComputeGaussWeights( dispersion_sq * dispersion_sq );

	return true;
}

//-----------------------------------------------------------------------
//! @brief 
//-----------------------------------------------------------------------
void SampleRenderer::OnRelease()
{
	mSaturn.Release();
	mGround.Release();
	ModelUtil::Release();

    for( uint32_t i=0; i<2; ++i )
    {
    	SAFE_RELEASE( mpSRV[ i ] );
    	SAFE_RELEASE( mpSmp[ i ] );
    	SAFE_RELEASE( mpTex[ i ] );
    	SAFE_RELEASE( mpRTV[ i ] );
    }

	SAFE_RELEASE( mpQuadVB );

    SAFE_RELEASE( mpCBufGauss );

	mShaders.Release();
}

//-----------------------------------------------------------------------
//! @brief foCX̏
//-----------------------------------------------------------------------
bool SampleRenderer::OnCreateDevice
(
	ID3D11Device *pDevice,
	const DXGI_SURFACE_DESC *pDesc
)
{
	HRESULT hr = S_OK;

	// f[eBeB[
	if ( !ModelUtil::Initialize( pDevice ) )
	{
		ELOG( "Error : ModelUitl::Iniatialize() Failed." );
		return false;
	}
	
	// f̃obt@𐶐
	if ( !mSaturn.CreateBuffer( pDevice ) )
	{
		ELOG( "Error : ModelUtil::CreateBuffer() Failed." );
		return false;
	}
	if ( !mGround.CreateBuffer( pDevice ) )
	{
		ELOG( "Error : ModelUtil::CreateBuffer() Failed." );
		return false;
	}

	// VF[_̃RpCݒ
	ShaderCompileDesc shaderCompileDescs[] = {
		{ SHADER_TYPE_VERTEX, "VSFunc_Pass1", "vs_4_0" },
   		{ SHADER_TYPE_VERTEX, "VSFunc_Pass2", "vs_4_0" },
		{ SHADER_TYPE_PIXEL,  "PSFunc_Pass1", "ps_4_0" },
   		{ SHADER_TYPE_PIXEL,  "PSFunc_Pass2", "ps_4_0" },
	};
	const uint32_t numShaderCompileDescs = sizeof( shaderCompileDescs ) / sizeof( shaderCompileDescs[ 0 ] );

	// t@CVF[_RpC
	if ( !mShaders.CreateFromFile( 
		pDevice,
		TEXT( "Media/Shaders/GaussianFilter.hlsl" ),
		shaderCompileDescs,
		numShaderCompileDescs ) )
	{
		ELOG( "Error : ShaderContainer::CreateFromFile() Failed." );
		return false;
	}

	// 2eNX`̐ݒ
	D3D11_TEXTURE2D_DESC texDesc;
	memset( &texDesc, 0, sizeof( texDesc ) );
	texDesc.Usage              = D3D11_USAGE_DEFAULT;
	texDesc.Format             = DXGI_FORMAT_R8G8B8A8_TYPELESS;
	texDesc.BindFlags          = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
	texDesc.Width              = pDesc->Width;
	texDesc.Height             = pDesc->Height;
	texDesc.CPUAccessFlags     = 0;
	texDesc.MipLevels          = 1;
	texDesc.ArraySize          = 1;
	texDesc.SampleDesc.Count   = 1;
	texDesc.SampleDesc.Quality = 0;

	// 2eNX`̐
    hr = pDevice->CreateTexture2D( &texDesc, NULL, &mpTex[ 0 ] );
	if ( FAILED( hr ) )
	{
		ELOG( "Error : ID3D11Device::CreateTexture2D() Failed." );
		return false;
	}
    hr = pDevice->CreateTexture2D( &texDesc, NULL, &mpTex[ 1 ] );
    if ( FAILED( hr ) )
    {
        ELOG( "Error : ID3D11Device::CreateTexture2D() Failed." );
        return false;
    }

	// _[^[Qbgr[̐ݒ
	D3D11_RENDER_TARGET_VIEW_DESC rtvDesc;
	memset( &rtvDesc, 0, sizeof( rtvDesc ) );
	rtvDesc.Format             = DXGI_FORMAT_R8G8B8A8_UNORM;
	rtvDesc.ViewDimension      = D3D11_RTV_DIMENSION_TEXTURE2D;

	// _[^[Qbgr[̐
	hr = pDevice->CreateRenderTargetView( mpTex[ 0 ], &rtvDesc, &mpRTV[ 0 ] );
	if ( FAILED( hr ) )
	{
		ELOG( "Error : ID3D11Device::CreateRenderTargetView() Failed." );
		return false;
	}
    hr = pDevice->CreateRenderTargetView( mpTex[ 1 ], &rtvDesc, &mpRTV[ 1 ] );
    if ( FAILED( hr ) )
    {
        ELOG( "Error : ID3D11Device::CreateRenderTargetView() Failed." );
        return false;
    }

	// VF[_\[Xr[̐ݒ
	D3D11_SHADER_RESOURCE_VIEW_DESC srvDesc;
	memset( &srvDesc, 0, sizeof( srvDesc ) );
	srvDesc.Format              = rtvDesc.Format;
	srvDesc.ViewDimension       = D3D11_SRV_DIMENSION_TEXTURE2D;
	srvDesc.Texture2D.MipLevels = 1;
	
	// VF[_\[Xr[̐
	hr = pDevice->CreateShaderResourceView( mpTex[ 0 ], &srvDesc, &mpSRV[ 0 ] );
	if ( FAILED( hr ) )
	{
		ELOG( "Error : ID3D11Device::CreateShaderResourceView() Failed." );
		return false;
	}
    hr = pDevice->CreateShaderResourceView( mpTex[ 1 ], &srvDesc, &mpSRV[ 1 ] );
    if ( FAILED( hr ) )
    {
        ELOG( "Error : ID3D11Device::CreateShaderResourceView() Failed." );
        return false;
    }

	// TvXe[g̐ݒ
	D3D11_SAMPLER_DESC smpDesc;
	memset( &smpDesc, 0, sizeof( smpDesc ) );
	smpDesc.Filter         = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
	smpDesc.AddressU       = D3D11_TEXTURE_ADDRESS_WRAP;
	smpDesc.AddressV       = D3D11_TEXTURE_ADDRESS_WRAP;
	smpDesc.AddressW       = D3D11_TEXTURE_ADDRESS_WRAP;
	smpDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
	smpDesc.MinLOD         = 0;
	smpDesc.MaxLOD         = D3D11_FLOAT32_MAX;

	// TvXe[g
	hr = pDevice->CreateSamplerState( &smpDesc, &mpSmp[ 0 ] );
	if ( FAILED( hr ) )
	{
		ELOG( "Error : ID3D11Device::CreateSamplerState() Failed." );
		return false;
	}
    hr = pDevice->CreateSamplerState( &smpDesc, &mpSmp[ 1 ] );
    if ( FAILED( hr ) )
    {
        ELOG( "Error : ID3D11Device::CreateSamplerState() Failed." );
        return false;
    }

	// ``p_
	struct QuadVertex
	{
		D3DXVECTOR3 Pos;
		D3DXVECTOR2 Uv;
	};

	// _f[^̐ݒ
	QuadVertex vertices[] = {
		{ D3DXVECTOR3( +1.0f, +1.0f, 0.0f ), D3DXVECTOR2( 1.0f, 0.0f ) },
		{ D3DXVECTOR3( +1.0f, -1.0f, 0.0f ), D3DXVECTOR2( 1.0f, 1.0f ) },
		{ D3DXVECTOR3( -1.0f, -1.0f, 0.0f ), D3DXVECTOR2( 0.0f, 1.0f ) },

   		{ D3DXVECTOR3( +1.0f, +1.0f, 0.0f ), D3DXVECTOR2( 1.0f, 0.0f ) },
		{ D3DXVECTOR3( -1.0f, -1.0f, 0.0f ), D3DXVECTOR2( 0.0f, 1.0f ) },
		{ D3DXVECTOR3( -1.0f, +1.0f, 0.0f ), D3DXVECTOR2( 0.0f, 0.0f ) },
	};

	// XgCh̐ݒ
	mQuadStride = sizeof( QuadVertex );

	// _obt@̐ݒ
	D3D11_BUFFER_DESC bufDesc;
	memset( &bufDesc, 0, sizeof( bufDesc ) );
	bufDesc.Usage          = D3D11_USAGE_DEFAULT;
	bufDesc.BindFlags      = D3D11_BIND_VERTEX_BUFFER;
	bufDesc.CPUAccessFlags = 0;
	bufDesc.ByteWidth      = sizeof( QuadVertex ) * 6;

	// Tu\[Xf[^̐ݒ
	D3D11_SUBRESOURCE_DATA initData;
	memset( &initData, 0, sizeof( initData ) );
	initData.pSysMem = vertices;

	// _obt@
	hr = pDevice->CreateBuffer( &bufDesc, &initData, &mpQuadVB );
	if ( FAILED( hr ) )
	{
		ELOG( "Error : ID3D11Device::CreateBuffer() Failed." );
		return false;
	}

    // KEXtB^p萔obt@̐ݒ
    memset( &bufDesc, 0, sizeof( bufDesc ) );
    bufDesc.Usage = D3D11_USAGE_DEFAULT;
    bufDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    bufDesc.CPUAccessFlags = 0;
    bufDesc.ByteWidth = sizeof( GaussBuffer );

    // KEXtB^p萔obt@̐
    hr = pDevice->CreateBuffer( &bufDesc, NULL, &mpCBufGauss );
    if ( FAILED( hr ) )
    {
        ELOG( "Error : ID3D11Device::CreatBuffer() Failed." );
        return false;
    }
	
	return true;
}

//-----------------------------------------------------------------------
//! @brief foCXj̏
//-----------------------------------------------------------------------
void SampleRenderer::OnDestroyDevice()
{
	mSaturn.ReleaseBuffer();
	mGround.ReleaseBuffer();
	ModelUtil::Release();

    for( uint32_t i=0; i<2; ++i )
    {
    	SAFE_RELEASE( mpSRV[ i ] );
    	SAFE_RELEASE( mpSmp[ i ] );
    	SAFE_RELEASE( mpTex[ i ] );
    	SAFE_RELEASE( mpRTV[ i ] );
    }

	SAFE_RELEASE( mpQuadVB );
    SAFE_RELEASE( mpCBufGauss );

	mShaders.Release();
}

//-----------------------------------------------------------------------
//! @brief Xbv`FC̃TCY̏
//-----------------------------------------------------------------------
bool SampleRenderer::OnResizeSwapChain
(
	ID3D11Device *pDeivce,
	IDXGISwapChain *pSwapChain,
	const DXGI_SURFACE_DESC *pDesc 
)
{
	D3DXMatrixPerspectiveFovLH( &mProjection, Math::PiOver4, DemoApp::GetInstance()->GetAspectRatio(), 1.0f, 1000.0f );
	return true;
}

//-----------------------------------------------------------------------
//! @brief XV
//-----------------------------------------------------------------------
void SampleRenderer::OnUpdate( double time, float elapsedTime )
{
	/* DO_NOTHING */
}

//-----------------------------------------------------------------------
//! @brief `揈
//-----------------------------------------------------------------------
void SampleRenderer::OnRender
(
	ID3D11Device* pDevice, 
	ID3D11DeviceContext* pImmediateContext, 
	double time,
	float  elapsedTime
)
{
	//===================================================================
	// eNX`ɕ`
	//===================================================================

	// ftHg̐[xEXeVr[擾
	ID3D11DepthStencilView* pDSV = DemoApp::GetInstance()->GetDefaultDepthStencilView();

	// Õ_[^[Qbgr[ɐ؂ւ
	pImmediateContext->OMSetRenderTargets( 1, &mpRTV[ 0 ], pDSV );
	pImmediateContext->ClearRenderTargetView( mpRTV[ 0 ], DemoApp::GetInstance()->GetDefaultClearColor() );
	pImmediateContext->ClearDepthStencilView( pDSV, D3D11_CLEAR_DEPTH, 1.0f, 0 );
	
	// ftHgVF[_擾
	ID3D11VertexShader *pVS = ModelUtil::GetDefaultVertexShader();
	ID3D11PixelShader  *pPS = ModelUtil::GetDefaultPixelShader();

	// VF[_ݒ
	pImmediateContext->VSSetShader( pVS, NULL, 0 );
	pImmediateContext->PSSetShader( pPS, NULL, 0 );

	// 萔obt@ݒ
	D3DXMatrixIdentity( &mWorld );
	ModelUtil::SetMatrix( pImmediateContext, mWorld, mView, mProjection );
	ModelUtil::SetVector( pImmediateContext, mLightDir, mViewDir );

	// f`
	mSaturn.Draw( pImmediateContext );

	// sړsvZ
	D3DXMATRIX translate;
	D3DXMatrixTranslation( &translate, 0.0f, -1.0f, 0.0f );

	// [hsݒ
	mWorld = translate;

	// s]
	ModelUtil::SetMatrix( pImmediateContext, mWorld, mView, mProjection );
	
	// nʂ`
	mGround.Draw( pImmediateContext );


	//===================================================================
	// Pass1 : ɂڂ
	//===================================================================

	// Õ_[^[Qbgr[ɐ؂ւ
	pImmediateContext->OMSetRenderTargets( 1, &mpRTV[ 1 ], pDSV );
	pImmediateContext->ClearRenderTargetView( mpRTV[ 1 ], DemoApp::GetInstance()->GetDefaultClearColor() );
	pImmediateContext->ClearDepthStencilView( pDSV, D3D11_CLEAR_DEPTH, 1.0f, 0 );

	// fobO`VF[_ݒ
	mShaders.SetShader( pImmediateContext, 0, 0 );

	// _OeNX`ݒ
    pImmediateContext->PSSetShaderResources( 0, 1, &mpSRV[ 0 ] );
    pImmediateContext->PSSetSamplers( 0, 1, &mpSmp[ 0 ] );

    // KEXtB^pp[^ݒ
    GaussBuffer buffer;
    memcpy( buffer.Weights, mWeights, sizeof( float ) * 8 );
    buffer.Offset.x = 16.0f / DemoApp::GetInstance()->GetWidth();
    buffer.Offset.y = 16.0f / DemoApp::GetInstance()->GetHeight();
    buffer.Size.x = (float)DemoApp::GetInstance()->GetWidth();
    buffer.Size.y = (float)DemoApp::GetInstance()->GetHeight();

    // 萔obt@ݒ
    pImmediateContext->UpdateSubresource( mpCBufGauss, 0, NULL, &buffer, 0, 0 );
    pImmediateContext->VSSetConstantBuffers( 0, 1, &mpCBufGauss );
    pImmediateContext->PSSetConstantBuffers( 0, 1, &mpCBufGauss );

	// _obt@ݒ
	uint32_t offset = 0;
	pImmediateContext->IASetVertexBuffers( 0, 1, &mpQuadVB, &mQuadStride, &offset );
	pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );

	// ``
	pImmediateContext->Draw( 6, 0 );


    //===================================================================
	// Pass2 : cɂڂ
	//===================================================================
    ID3D11RenderTargetView* pRTV = DemoApp::GetInstance()->GetDefaultRenderTargetView();
	
   	// Õ_[^[Qbgr[ɐ؂ւ
	pImmediateContext->OMSetRenderTargets( 1, &pRTV, pDSV );
	pImmediateContext->ClearRenderTargetView( pRTV, DemoApp::GetInstance()->GetDefaultClearColor() );
	pImmediateContext->ClearDepthStencilView( pDSV, D3D11_CLEAR_DEPTH, 1.0f, 0 );

	// fobO`VF[_ݒ
	mShaders.SetShader( pImmediateContext, 1, 1 );

	// _OeNX`ݒ
	pImmediateContext->PSSetShaderResources( 0, 1, &mpSRV[ 1 ] );
	pImmediateContext->PSSetSamplers( 0, 1, &mpSmp[ 1 ] );

    // 萔obt@ݒ
    pImmediateContext->UpdateSubresource( mpCBufGauss, 0, NULL, &buffer, 0, 0 );
    pImmediateContext->VSSetConstantBuffers( 0, 1, &mpCBufGauss );
    pImmediateContext->PSSetConstantBuffers( 0, 1, &mpCBufGauss );

	// _obt@ݒ
	offset = 0;
	pImmediateContext->IASetVertexBuffers( 0, 1, &mpQuadVB, &mQuadStride, &offset );
	pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );

	// ``
	pImmediateContext->Draw( 6, 0 );

}

//-----------------------------------------------------------------------
//! @brief KEX̏d݂vZ
//-----------------------------------------------------------------------
void SampleRenderer::ComputeGaussWeights( float dispersion )
{
    float total = 0.0f;
    const int NUM_WEIGHT = 8;

    for( int i=0; i<NUM_WEIGHT; ++i )
    {
        float pos = 1.0f + 2.0f * (float)i;
        mWeights[ i ] = expf( -0.5f * ( pos * pos ) / dispersion );
        total += 2.0f * mWeights[ i ];
    }

    float invTotal = 1.0f / total;
    for( int i=0; i<NUM_WEIGHT; ++i )
    {
        mWeights[ i ] *= invTotal;
    }
}

} // namespace SampleProgram
} // namespace Asura
